# -*- coding: utf-8 -*-
__all__ = ['Server', 'AllMoves', ]
__author__ = "Obsession <2608135678@qq.com>"

import random, asyncio, pprint, logging
import zmq, pickle, multiprocessing
log = lambda *j:logging.getLogger(__name__).warning(
        '   '.join([pprint.pformat(i, width=150) for i in j]))


# 开局子力分布图 (网格:棋子)
SPOT_PIECE = {
        51:  36, 52:  35, 53:  34, 54:  33, 55:  32, 56:  33, 57:  34, 58:  21, 59:  36,
        67:  0,  68:  0,  69:  0,  70:  0,  71:  0,  72:  0,  73:  0,  74:  0,  75:  0,
        83:  0,  84:  37, 85:  0,  86:  0,  87:  0,  88:  0,  89:  0,  90:  37, 91:  0,
        99:  38, 100: 0,  101: 38, 102: 0,  103: 38, 104: 0,  105: 38, 106: 0,  107: 38,
        115: 0,  116: 0,  117: 0,  118: 0,  119: 0,  120: 0,  121: 0,  122: 0,  123: 0,
        131: 0,  132: 0,  133: 0,  134: 0,  135: 0,  136: 0,  137: 0,  138: 0,  139: 0,
        147: 22, 148: 0,  149: 22, 150: 0,  151: 22, 152: 0,  153: 22, 154: 0,  155: 22,
        163: 0,  164: 21, 165: 0,  166: 0,  167: 0,  168: 0,  169: 0,  170: 0,  171: 0,
        179: 0,  180: 0,  181: 0,  182: 0,  183: 0,  184: 0,  185: 0,  186: 0,  187: 0,
        195: 20, 196: 19, 197: 18, 198: 17, 199: 16, 200: 17, 201: 18, 202: 19, 203: 20.
}

async def get_all_route(we, spot_piece):
    routes = []
    for spot,piece in spot_piece.items():
        if not piece & we:continue  # 忽略对方棋子
        for seat in AllMoves.getRoute(spot, piece, spot_piece):
            routes.append([spot, piece, seat])
    return routes



class InterfaceEvaluation(object):
    """ 界面评估, 不会写了, 先返回个随机分数吧 """
    def __init__(self, our, spot_piece):
        self.our = our
        self.spot_piece = spot_piece

    async def scoring(self, go_spot, go_piece, to_spot):
        # print(go_spot, go_piece, to_spot, to_piece)
        return random.randrange(-10000, 10000)


class Server(multiprocessing.Process):
    """ 电脑生成应招, 接收2个参数, 1.执红'red'或黑'black', 2当前子力分布图
        返回结果 (从网格, 棋子, 到网格)]
    """
    def __init__(self):
        super().__init__()
        self.__socket = None
        self.our = 'black'
        self.spot_piece = SPOT_PIECE
        self.red_black = {'red':16, 'black':32}

    async def score(self, route):
        assesslog = InterfaceEvaluation(self.our, self.spot_piece)
        return route, await assesslog.scoring(*route)

    async def handle(self, task):
        """ 先生成所有着法, 异步方式对每步进行界面评估,
            最后返回界面得分最高的应招 (从网格, 棋子, 到网格)
        """
        try:
            self.our, self.spot_piece = task # 获取参数
        except Exception:
            return print('服务参数错误, 只接收如: ["red", {}]', task)
        we = self.red_black[self.our]
        all_route = await get_all_route(we, self.spot_piece)  # 所有走法
        tasks = [self.score(x) for x in all_route] # 为每种走法创建一个线程
        result = await asyncio.gather(*tasks, return_exceptions=True)   # 异步评估
        optimum = sorted(result, key=lambda i:i[1], reverse=True)   # 得分排序
        return optimum and optimum[0][0]

    def run(self):
        context = zmq.Context()
        self.__socket = context.socket(zmq.REP)
        self.__socket.bind("tcp://*:6789")
        asyncio.run(self.eventPool())

    async def eventPool(self):
        while True:
            try:
                message = self.__socket.recv()
                task = pickle.loads(message)
                if task == 'stop':break
                result = await self.handle(task)
                self.__socket.send(pickle.dumps(result))
            except Exception as exc:
                logging.exception(f"服务器异常: {exc}", exc_info=True)
                break
        print('服务器退出 ...')


class AllMoves(object):
    chess_state = 1
    cmp = lambda x,y: x < y and -1 or (x > y and 1 or 0)
    nine = [54, 55, 56,
            70, 71, 72,
            86, 87, 88,
            166, 167, 168,
            182, 183, 184,
            198, 199, 200]   # 九宫

    def __init__(self, ):
        self.piece_name = dict(zip(
                [16,17,18,19,20,21,22,32,33,34,35,36,37,38],
                "帅仕相馬車炮兵将士象馬車炮卒"))

    @classmethod
    def getRoute(cls, spot, piece, situation):
        """ 产生所有能够走棋的位置 (spot=网格,piece=棋子,situation=子力分布图) """
        route = []
        if (piece in (16,32)): # 将帅
            route = [ (spot + i) for i in (-1, 1, -16, 16) if ((spot + i) in cls.nine) ]
            check = [i for i in ((spot<128 and cls.nine[9:]) or cls.nine[:9]) if situation[i] in (16, 32)][0]
            x = (((cls.cmp(spot, check) > 0) and (((spot // 16) - (check // 16)), 16)) or (((check // 16) - (spot // 16)), -16))
            j = ((spot % 16) - (check % 16))
            if j in (1,  -1):
                if (not [ i for i in range(1, x[0]) if situation[((i * x[1]) + check)] ]):
                    route.pop(route.index((spot - j)))
            elif (not j):
                for i in range(1, int(x[0] + 1)):
                    j = ((i * x[1]) + check)
                    if (j == spot):
                        route = [check]
                    elif ((i == (x[0] - 1)) and situation[j]):
                        route.pop(route.index(j))
                        break
                    elif situation[j]:
                        break
        elif (piece in (17, 33)):   # 士
            route = [ (spot + i) for i in (-15, 15, -17, 17) if ((spot + i) in cls.nine) ]
        elif (piece in (18, 34)):  # 象
            route = [ (spot + (i * 2)) for i in (-15, 15, -17, 17) if (((spot + i) in situation)
                    and ((not situation[(spot + i)]) and (((spot + (i * 2)) // 16) not in (6, 9)))) ]
        elif (piece in (19, 35)):  # 马
            route = [ (spot + j) for (i, j,) in ((-1, -18), (-1, 14), (1, -14), (1, 18), (-16, -31), (-16, -33), (16, 31), (16, 33)) if (((spot + i) in situation) and (not situation[(spot + i)])) ]
        elif piece in (20, 36):     # 車
            for i in (-1, 1, -16, 16):
                for j in range(1, 10):
                    j = ((i * j) + spot)
                    route.append(j)
                    if ((j not in situation) or situation[j]):
                        break
        elif (piece in (21, 37)):  # 炮
            for i in (-1, 1, -16, 16):
                x = False
                for j in range(1, 10):
                    j = ((i * j) + spot)
                    if ((j not in situation) or (x and situation[j])):
                        route.append(j)
                        break
                    elif situation[j]:
                        x = True
                    elif (not x):
                        route.append(j)
        elif (piece in (22, 38)):  # 兵卒
            route = ((((piece - 6) in [ situation[i] for i in cls.nine[9:] ]) and (((spot < 128) and [ (spot + i) for i in (-1, 1, -16) ]) or [(spot - 16)])) or (((spot > 128) and [ (spot + i) for i in (-1, 1, 16) ]) or [(spot + 16)]))
        route = [ i for i in route if ((i in situation) and (((piece & 16) != (situation[i] & 16)) or ((piece & 32) != (situation[i] & 32)))) ]
        return route


